/***************************************************************************
                          game_client.cpp  -  description
                             -------------------
    begin                : Thu Feb 3 2000
    copyright            : (C) 2000 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/


#include "game_client.h"
#include "all.h"

//-----------------------------------------------------------------------------
Game_client::Game_client()
{
	last_tick_time = 0;
	last_tick_duration = 1;
	loader = NULL;
	dropshadow = true;
  viewport = new Viewport(320,240);
	setgeom(0,0,320,240,2);
	bars_at_top = false;

	elevate_up = false;
	elevate_down = false;
}

//-----------------------------------------------------------------------------
Game_client::~Game_client()
{
	if (tcp_client.send(NETSTATE::QUIT) < 0)
		logmsg(lt_error,"unable to tell the server we want to quit.");

	if (viewport) delete viewport;
}

//-----------------------------------------------------------------------------
//call to get things going:
int Game_client::init(Loader *loader_, Soundsystem *soundsystem_, const char *player_name_)
{
	loader = loader_;
	soundsystem = soundsystem_;
	player_name = (char *)player_name_;

	health_bar.setgeom(5,viewport->get_ysize()-(7+5),100,7);
	health_bar.bar_color = PIXELA(128,0,255,8);
	health_bar.render();

	return 0;
}

//-----------------------------------------------------------------------------
int Game_client::setgeom(int xpos_, int ypos_, int view_xsize, int view_ysize, int view_quality_)
{
	xpos = xpos_;
	ypos = ypos_;
	view_quality = view_quality_;
	if (viewport) delete viewport;
  viewport = new Viewport(view_xsize, view_ysize);
	return 0;
}

//-----------------------------------------------------------------------------
int Game_client::join_game(const char *server, int server_port)
{
	{
		unsigned int firsttry = time_ms();
		int error = -1;

		logmsg(lt_debug,"Trying to connect to server %s:%i",server,server_port);
		while (error < 0 && time_elapsed(firsttry,time_ms()) < 10000) {
			error = tcp_client.connect(server,server_port);
			if (error < 0) {
				sleep(1);
				logmsg(lt_debug,"Couln't connect, but trying again...");
			}
		}

		if (error < 0) {
			logmsg(lt_error,"Unable to get connected to the server, quitting. %i",error);
			return -1;
		} else
			logmsg(lt_message,"Yes!, we are connected to the server!");
	}

	{
		Net_server_signature server_signature;
		if (tcp_client.recv(&server_signature) < 0) {
			logmsg(lt_error,"Unable to get server signature");	
			return -2;
		}

    if (!server_signature.server_ok()) {
			logmsg(lt_fatal,"Server is not a NiL server, it said: %s",server_signature.signature);
			return -3;
		};

    if (!server_signature.version_ok()) {
			logmsg(lt_fatal,"Server is a NiL server, but it has the wrong version: %s you are using %s",server_signature.version,PROTOCOL_VERSION);
			return -4;
		};
	}

	//send join request
	Net_joinrequest joinrequest;
	joinrequest.color = PIXEL(255,128,0);
	strcpy(joinrequest.name,player_name);
  if (tcp_client.send(&joinrequest) < 0) {
		logmsg(lt_error,"tcp_client.send(&joindata,sizeof(joindata) failed.");
		return -5;
	}
	logmsg(lt_debug,"join request sent");

	//get join reply
	Net_joinreply joinreply;
  if (tcp_client.recv(&joinreply) < 0) {
		logmsg(lt_error,"tcp_client.recv(&joinreply,sizeof(joinreply)) failed.");
		return -6;
	}
	logmsg(lt_debug,"got join reply from server.");				

	//create client world
//	Clientworld clientworld(&loader,joinreply.objectid);
	clientworld.init(loader, soundsystem, joinreply.objectid);	
	logmsg(lt_debug,"client world created, about to initmap.");				

	{
		//get the compressed map:
		char *map_data = (char*)malloc(joinreply.compressed_size);
 	 	if (tcp_client.recv(map_data,joinreply.compressed_size) < 0) {
			logmsg(lt_error,"Couldn't read compressed map diff.");
			return -7;
		}
		logmsg(lt_debug,"got map.");				

		//have the clientworld unpack the packed gob:
		clientworld.createmap(joinreply,map_data);
		free(map_data);
		logmsg(lt_debug,"initmap done.");
  }

	viewport->setmapsize(joinreply.map_xsize,joinreply.map_ysize);

	//send handshake:
	int handshake = 42;
  if (tcp_client.send(&handshake,sizeof(handshake)) < 0) {
		logmsg(lt_error,"Couldn't send handshake.");
		return -8;
	}
	logmsg(lt_debug,"ready to go.");
	return 0;
}


//-----------------------------------------------------------------------------
//void Game_client::get_join_status(char *text, int textsize, int &progress_, int &max_progress_); remember to have a semaphor

//-----------------------------------------------------------------------------
//poll once for each frame:
int Game_client::think()
{
	int netstate = -1;
	int this_update_id = -1;

	//Get the update id:
	tcp_client.recv(netstate);
	if (netstate != NETSTATE::UPDATEID) {
		logmsg(lt_error,"protocol error 1.");
		return -1;
	}
	if (tcp_client.recv(this_update_id) <0) {
		logmsg(lt_error,"unable to get the update_id from the server.");
	}

	//Get the world update itself
	tcp_client.recv(netstate);
	if (netstate != NETSTATE::WORLDUPDATE) {
		logmsg(lt_error,"protocol error 2.");
		return -1;
	}
	if (tcp_client.recv(&clientworld.serializer_writer) < 0) {
		logmsg(lt_error,"unable to get the worldupdate from the server.");
		return -1;
	}

	//Get the done flag:
	tcp_client.recv(netstate);
	if (netstate != NETSTATE::DONE) {
		logmsg(lt_error,"protocol error 3.");
		return -1;
	}


	//Send ack for this update:
	if (tcp_client.send(NETSTATE::UPDATEID) < 0) {
		logmsg(lt_error,"unable to send state to the server.");
		return -2;  	
	}
	if (tcp_client.send(this_update_id) < 0) {
		logmsg(lt_error,"unable to send this_update_id to the server.");
		return -2;  	
	}

	//send the current controls
	if (tcp_client.send(NETSTATE::CONTROLS) < 0) {
		logmsg(lt_error,"unable to send state to the server.");
		return -2;  	
	}
	if (tcp_client.send(&controls) < 0) {
		logmsg(lt_error,"unable to send controls to the server.");
		return -2;
	}

	//Tell the server that there is no more lumps for now
	if (tcp_client.send(NETSTATE::DONE) < 0) {
		logmsg(lt_error,"unable to send state to the server.");
		return -2;  	
	}


  //update some statistics:
	last_update_size = clientworld.serializer_writer.getlength();
	last_tick_duration = clientworld.gettime()-last_tick_time;
	last_tick_time = clientworld.gettime();
	if (last_tick_duration > 10) {
		last_tick_duration = 1;
		trend_tick_duration = 1;
	}
	trend_tick_duration += (last_tick_duration-trend_tick_duration)/20;


	clientworld.client_think();

	controls_tick();

	return 0;
}

//-----------------------------------------------------------------------------
//call to have this game client draw it's view of the world
int Game_client::draw(Raw_surface *target)
{
//	Obj_player_avatar *avatar = clientworld.get_player_avatar(clientworld.getowner());
	Obj_player_data *player_data	= (Obj_player_data *)clientworld.get_object(clientworld.getowner(),ot_player_data);
	if (!player_data) {
  	logmsg(lt_warning,"Unable to find player_data");
		return -1;
	}

	Obj_player_avatar *avatar 		= player_data->get_player_avatar();
	//figure out if we are to put the bars at the top or the bottom:
	if (avatar) {
		float x,y;
		avatar->getpos(x,y);
		int switch_margin = viewport->get_ysize()/2;

    if (bars_at_top) {

    	if (y < switch_margin)
				bars_at_top = false;

		} else {

    	if (y > clientworld.map_buffer.get_ysize()-switch_margin)
				bars_at_top = true;

		}

	}

	//Figure out the pixel position of the bars
	int bars_ypos = bars_at_top?5:viewport->get_ysize()-(7+5);				

  //Update the health bar:
	if (avatar) health_bar.setpos((avatar->gethealth()*1.0)/MAX_HEALTH);
	health_bar.ypos = bars_ypos;

	//Update the weapon bars:

	if (player_data) {
 		int bar_width = 25;
		int bar_start = 120;
		for (int s=0;s<WEAPON_SLOTS;s++) {
      int this_bar_width=controls.get_weapon()==s?ROUND(bar_width*2):bar_width;
      weapon_bar[s].setgeom(bar_start,bars_ypos,this_bar_width,7);

			Obj_weapon *weapon = player_data->get_weapon(s);
			if (weapon) {

				//Figure out the color of the bar
				int32 newcolor = weapon->canfire()?PIXELA(0,255,0,6):PIXELA(255,0,0,6);
				if (weapon_bar[s].bar_color != newcolor) {
					weapon_bar[s].bar_color = newcolor;
					weapon_bar[s].render();
				}
				
				float readyness = weapon->getreadyness();
				weapon_bar[s].setpos(readyness);
			} else {
				weapon_bar[s].setpos(1);
				logmsg(lt_warning,"Unable to find the weapon for slot %i",s);
			}

			bar_start += this_bar_width+10;		
		}
	}

	//figure out what to view
	int vx = 0;
	int vy = 0;
	clientworld.get_player_pos(clientworld.getowner(),vx,vy);
	viewport->addsample(vx,vy,last_tick_duration);

	clientworld.draw(viewport);

	if (dropshadow)
		viewport->drop_shadow();

	//draw the player specific UI bits:
	health_bar.draw(viewport);
	for (int s=0;s<WEAPON_SLOTS;s++)
    weapon_bar[s].draw(viewport);

	viewport->draw_copy_stretch(target,
		0,0,
		viewport->get_xsize(), viewport->get_ysize(),
		xpos,  ypos,
		view_quality
	);
	
	return 0;
}

//-----------------------------------------------------------------------------
void Game_client::controls_tick()
{
	if (elevate_down)
		controls.set_elevation(controls.get_elevation()-0.2);	
	if (elevate_up)
		controls.set_elevation(controls.get_elevation()+0.2);	
	controls_ticked = true;
}

//-----------------------------------------------------------------------------
bool Game_client::handle_key(Nil_keyevent *keyevent)
{
	if (!keyevent || !keyevent->sdl_keyevent) return false;

	SDLKey sdlkey    = keyevent->sdl_keyevent->keysym.sym;
	bool key_pressed = keyevent->sdl_keyevent->state == SDL_PRESSED;
	int function = keyevent->keyfunction & KEYFUNCTION::FUNCTION_MASK;

	switch (function) {

  	case (KEYFUNCTION::UP) :{
			if (key_pressed) {
				elevate_up = true;
				controls_ticked = false;
			} else {
				if (!controls_ticked) controls_tick();
				elevate_up = false;
			}
			return true;
		}

    case (KEYFUNCTION::DOWN) :{
			if (key_pressed) {
				elevate_down = true;
				controls_ticked = false;
			} else {
				if (!controls_ticked) controls_tick();
				elevate_down = false;
			}
			return true;
		}

    case (KEYFUNCTION::LEFT) :{
			controls.clear_flag(Controls::cf_direction_east);
			controls.set_flag(Controls::cf_walk,key_pressed);
			return true;
		}

		case (KEYFUNCTION::RIGHT) :{
			controls.set_flag(Controls::cf_direction_east);
			controls.set_flag(Controls::cf_walk,key_pressed);
			return true;
		}

		case (KEYFUNCTION::JUMP) :{
			controls.set_flag(Controls::cf_jump,key_pressed);
			return true;
		}

		case (KEYFUNCTION::HOOK_FIRE) :{
     	controls.set_flag(Controls::cf_hook_fire,key_pressed);
			return true;
		}

		case (KEYFUNCTION::HOOK_RELEASE) :{
			controls.set_flag(Controls::cf_hook_release,key_pressed);
			return true;
		}

		case (KEYFUNCTION::HOOK_IN) :{
			controls.set_flag(Controls::cf_hook_in,key_pressed);
			return true;
		}

		case (KEYFUNCTION::HOOK_OUT) :{
			controls.set_flag(Controls::cf_hook_out,key_pressed);
			return true;
		}

		case (KEYFUNCTION::FIRE) :{
			controls.set_flag(Controls::cf_shoot,key_pressed);
			return true;
		}

		case (KEYFUNCTION::DIG) :{
			controls.set_flag(Controls::cf_dig,key_pressed);
			return true;
		}

		case (KEYFUNCTION::WEAPON1) :{
			if (key_pressed)
       	controls.set_weapon(0);
			return true;
		}

		case (KEYFUNCTION::WEAPON2) :{
			if (key_pressed)
       	controls.set_weapon(1);
			return true;
		}

		case (KEYFUNCTION::WEAPON3) :{
			if (key_pressed)
       	controls.set_weapon(2);
			return true;
		}

		case (KEYFUNCTION::WEAPON4) :{
			if (key_pressed)
       	controls.set_weapon(3);
			return true;
		}

		case (KEYFUNCTION::WEAPON5) :{
			if (key_pressed)
       	controls.set_weapon(4);
			return true;
		}

    default: {
	   	logmsg(lt_error,"Unhandled key: 0x%x (%s) -> 0x%x",sdlkey,SDL_GetKeyName(sdlkey),function);
			return false;
		}

	} //switch (function) {

	return false;
}
