/***************************************************************************
                          loader.cpp  -  description
                             -------------------
    begin                : Sun Oct 17 1999
    copyright            : (C) 1999 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/

#include "systemheaders.h"
#include "loader.h"
#include "fluff.h"
#include "nil_files.h"
#include "fonts.h"

#include <list>
#ifdef HAS_STL_HASHMAP
#include <hash_map>
#else
#include <map>
#endif

//-----------------------------------------------------------------------
//The mmap() faker should probably be moved into a lib of it's own.
#ifdef _POSIX_MAPPED_FILES

#define nil_mmap mmap
#define nil_munmap munmap

#else

// *WARNING* Not using POSIX mapped files mmap(2)

//here we fake the mmap() routines, these should only be used on faulty OS'es!

/* returns pointer to the data or -1 in case of an error */
void *nil_mmap(void  *start, size_t length, int prot, int flags, int fd, off_t offset)
{
	void *data = malloc(length);
	if (!data) {
		logmsg(lt_error,"Unable to malloc() %i bytes for nil_mmap()",length);
		return MAP_FAILED;
	}
	if (read(fd, data, length) < 0) {
		free(data);
		logmsg(lt_error,"Unable to read() a file for nil_mmap()");
		return MAP_FAILED;
	}
	return data;
}

/* returns 0 or -1 in case of an error */
int nil_munmap(void *start, size_t length)
{
	if (start) {
		free(start);
	}
	return 0;
}

#endif //#else #ifdef USE_REAL_MMAP
//=======================================================================
//=======================================================================



//=======================================================================
//=======================================================================
//This should mmap the .lump file or make it look like it did so (via fopen()+fread())
//it should ALSO load all the data from the standalone files if it is just a directory
//all the files (from .lump or directory) should result in a Loader_file entry in the
//Loaders file map
//-----------------------------------------------------------------------
class Loader_files;
class Loader_source {
public:
	Loader_source();
	virtual ~Loader_source();

	const char *get_loaded_data() {return loaded_data;}
protected:
	Loader_files *files;
	bool initialized;
	char *loaded_data;
};

//-----------------------------------------------------------------------
class Loader_sources: public std::list<Loader_source *>{};

//-----------------------------------------------------------------------
class Loader_file {
public:
	typedef enum Mem_cleanup {
		MC_NONE, //The data source will clean up
		MC_FREE, //free your offset
		MC_UNMAP //unmap your offset
	};

	Loader_file(char *name_, int offset_, Loader_source *source_, int size_, Mem_cleanup mem_cleanup_);
	~Loader_file();

	const char *getname() {return name;}
	int getsize() {return size;}

	//does a traditional copy:
	bool fileread(void *target, int bytes, int pos=0);

	//returns a pointer to the already loaded data:
	//extreme care must be taken when meddeling with this!
	//you should NEVER write to data gotten from here and great
	//care must be taken to not read too far.
	const void *getdata();

	//used for backtracking.
	Loader_source* getsource() {return source;}

protected:
	Mem_cleanup mem_cleanup;
	char *name;
	Loader_source *source;
	int offset;	//offset from source->loaded_data to the start of the file
	int size;
};

//-----------------------------------------------------------------------
//The single global place to store the global directory:
#ifdef HAS_STL_HASHMAP
struct Streq {
	bool operator()(const char* s1, const char* s2) const
  {
		return strcmp(s1, s2) == 0;
	}
};

class Loader_files  : public std::hash_map<const char *,Loader_file *,hash<const char*>,Streq>{};

#else

struct Strless {
	bool operator()(const char* s1, const char* s2) const
  {
		return strcmp(s1, s2) < 0;
	}
};

class Loader_files  : public std::map<const char *,Loader_file *,Strless>{};

#endif

//specializations of the Loader source, use for actually loading data:

//-----------------------------------------------------------------------
class Loader_source_lump :public Loader_source {
public:
	Loader_source_lump();
	virtual ~Loader_source_lump();

	bool load_lump(Loader_files *files_, char *file_name);
protected:
	int data_length;
};

//-----------------------------------------------------------------------
class Loader_source_directory :public Loader_source {
public:
	Loader_source_directory();
	virtual ~Loader_source_directory();

	bool load_directory(Loader_files *files_, char *root_path);

protected:
	int data_length;
	Loader_files *files;
	bool recurse_dir(const char *root, const char *path);
	bool add_file(size_t size, char *fullpath, char *filename);
};

//============================================================================
//implementation of the helper classes:

//============================================================================
//class Loader_source {
Loader_source::Loader_source()
{
	initialized = false;
	files = NULL;
	loaded_data = NULL;
}

//-----------------------------------------------------------------------
Loader_source::~Loader_source()
{
	/*
	  This is the default destructor for data sources,
	  it removes the all this files that came from this source
	*/

	if (!files) return;
/*
	start_over:
	Loader_files::iterator i = files->begin();
	while (i != files->end()) {
	
		if (i->second->getsource() == this) {
			Loader_file *oldfile = i->second;
			files->erase(i);
			delete oldfile;

			//start over so we have a valid iterator (all elements may be moved around by the erase)
			goto start_over;
		}

		i++;
	}
*/
}

//============================================================================
//class Loader_file {
Loader_file::Loader_file(char *name_, int offset_, Loader_source *source_, int size_, Mem_cleanup mem_cleanup_)
{
	name = strdup(name_);
	offset = offset_;
	source = source_;
	size = size_;
	mem_cleanup = mem_cleanup_;
}

//-----------------------------------------------------------------------
Loader_file::~Loader_file()
{
	switch (mem_cleanup) {
		case MC_NONE : {
			//nothing to do.
			break;
		}

	  	case MC_FREE : {
			free((void *)offset);
			break;
		}

		case MC_UNMAP : {
			nil_munmap((void *)offset,size);
			break;
		}

		default : {
			int i = mem_cleanup;
			logmsg(lt_error,"Invalid mem_cleanup in ~Loader_file(): %i",i);
		}
	}

	free(name);
}

//-----------------------------------------------------------------------
//does a traditional copy:
bool Loader_file::fileread(void *target, int bytes, int pos)
{
	if (pos + bytes > size) {
		logmsg(lt_error,"Someone tried to read %i bytes more than possible from the file %s",(pos+bytes)-size,name);
		return false;
	}

	memcpy(target,((char *)getdata())+pos,bytes);

	return true;
}

//-----------------------------------------------------------------------
const void *Loader_file::getdata()
{
	return source->get_loaded_data()+offset;
}

//============================================================================
//specializations of the Loader source, use for actually loading data:

//============================================================================
//class Loader_source_lump :public Loader_source {
//-----------------------------------------------------------------------
Loader_source_lump::Loader_source_lump() : Loader_source()
{
	data_length = 0;
}

//-----------------------------------------------------------------------
Loader_source_lump::~Loader_source_lump()
{	
	if (data_length && initialized) {
		nil_munmap(loaded_data, data_length);
	}
}

//-----------------------------------------------------------------------
bool Loader_source_lump::load_lump(Loader_files *files_, char *file_name)
{
	if (initialized) return false;
	files = files_;

	//get size of file and put it in data_length
	struct stat stats;
	if (stat(file_name, &stats) < 0) {
		logmsg(lt_error,"Unable to get the size of the .lump file %s",file_name);
		return false;
	}
	data_length = stats.st_size;

	//mmap the file:
	int fd = open(file_name, O_RDONLY);
	if (fd < 0) {
		logmsg(lt_error,"Unable to open the file %s as read only: %i",file_name,fd);
		return false;
	}
	loaded_data = (char *)nil_mmap(NULL, stats.st_size, PROT_READ, MAP_SHARED, fd, 0);
	close(fd);
	if (loaded_data == MAP_FAILED) {
		logmsg(lt_error,"Unable to mmap() the file %s as read only: %i",file_name,errno);
		return false;
	}

	//walk the file to find all dir entries and store them in Loader_file objects	in files
	char *data = (char*)loaded_data;

	Lump_header header;
	memcpy(&header,data,sizeof(header));
	header.swap();
	if (header.magic != LUMP_HEADER_MAGIC) {
		logmsg(lt_error,"The magic of the lumpfile %s was 0x%x and it should have been 0x%x",file_name,header.magic,LUMP_HEADER_MAGIC);
		return false;
	}

	//process the directory:
	Lump_directory_header directory;
	memcpy(&directory,data+header.directory_offset,sizeof(directory));
	directory.swap();
	if (directory.magic != LUMP_DIRECTORY_HEADER_MAGIC) {
		logmsg(lt_error,"The magic of the directory in %s was 0x%x and it should have been 0x%x", file_name, directory.magic, LUMP_DIRECTORY_HEADER_MAGIC);
		return false;
	}

	int entry_offset = header.directory_offset + sizeof(directory);
	for (int i=0;i<directory.items;i++) {

		//get the item
		Lump_directory_item item;
		memcpy(&item,data+entry_offset,sizeof(item));
		item.swap();
		if (item.magic != LUMP_DIRECTORY_ITEM_MAGIC) {
			logmsg(lt_error,"The magic of directory item #%i in %s was 0x%x and it should have been 0x%x", i, file_name, item.magic, LUMP_DIRECTORY_ITEM_MAGIC);
			return false;
		}
		
		//check to see if we already have that file:
		Loader_files::iterator existing_file = files->find(file_name);
		if (existing_file != files->end()) {
			logmsg(lt_error,"File ignored, already available from another source: %s",file_name);
		} else {

			//save info about the new file:
			Loader_file *newfile = new Loader_file(
				data+entry_offset+sizeof(item), item.file_offset, this, item.file_length, Loader_file::MC_NONE
			);

			files->insert(std::pair<const char*,Loader_file *>(newfile->getname(),newfile));
		}

		//advance to next entry:
		entry_offset += sizeof(item) + item.name_length;
	}		

	initialized = true;

	return true;
}

//============================================================================
//class Loader_source_directory :public Loader_source {
Loader_source_directory::Loader_source_directory() : Loader_source()
{
	files = NULL;
	loaded_data = NULL;
}

//-----------------------------------------------------------------------
Loader_source_directory::~Loader_source_directory()
{
	if (loaded_data) {
		free(loaded_data);
	}
}

//-----------------------------------------------------------------------
bool Loader_source_directory::load_directory(Loader_files *files_, char *root_path)
{
	if (initialized) return false;

	files = files_;
	if (recurse_dir(root_path,"")) {
		initialized = true;
		return true;
	}
	return false;
}

//-----------------------------------------------------------------------
bool Loader_source_directory::recurse_dir(const char *root, const char *path)
{
	char *thisdir = (char*)malloc(strlen(root)+strlen(path)+2);
	strcpy(thisdir, root);
	strcat(thisdir, "/");
	strcat(thisdir, path);

	DIR *dir = opendir(thisdir);
	struct dirent *curentry;
	while ((curentry = readdir(dir)) != 0) {

		//partial path to current file
		char *curpath = (char*)malloc(strlen(path)+strlen(curentry->d_name)+2);
		strcpy(curpath, path);
		strcat(curpath, curentry->d_name);

		//full path to current file:
		char *curfile = (char*)malloc(strlen(root)+strlen(curpath)+2);
		strcpy(curfile, root);
		strcat(curfile, "/");
		strcat(curfile, curpath);

#ifdef _WIN32
		// Windows file names aren't case sensitive, so we'll just assume all
		// file names were lowercase.
		_strlwr(curpath);
		_strlwr(curfile);
#endif

		struct stat stats;
		if (stat(curfile, &stats) == 0) {

			//this is the leaf case:
			if (S_ISREG(stats.st_mode)) {
				if (!add_file(stats.st_size, curfile, curpath))
					return false;
			}
			
			//complex case, so we recurse
			if (S_ISDIR(stats.st_mode) && *curentry->d_name != '.') {
				strcat(curpath, "/");
				if (!recurse_dir(root, curpath))
					return false;
			}
		}

		free(curfile);
		free(curpath);
	}
	closedir(dir);
	free(thisdir);
	return true;
}

//-----------------------------------------------------------------------
bool Loader_source_directory::add_file(size_t size, char *fullpath, char *filename)
{
	FILE *fp;
	size_t count, padding, offset;

	Loader_files::iterator existing_file = files->find(filename);
	if (existing_file != files->end()) {
		logmsg(lt_error,"File ignored, already available from another source: %s",filename);
		return true;
	}

	padding = (data_length + sizeof(long) - 1) & (sizeof(long));
	if (loaded_data) {
		char *data;
		data = (char*)realloc(loaded_data, data_length+padding+size);
		if (!data) {
			return false;
		}
		loaded_data = data;
		offset = data_length + padding;
		data_length += padding + size;
	} else {
		loaded_data = (char*)malloc(size);
		if (!loaded_data) {
			return false;
		}
		offset = 0;
		data_length = size;
	}

	fp = fopen(fullpath, "rb");
	if (!fp) {
		logmsg(lt_error,"Unable to open the file %s as read only", fullpath);
		//allocated mem not freed
		return false;
	}
	count = fread(loaded_data + offset, 1, size, fp);
	fclose(fp);
	if (count < size) {
		logmsg(lt_error,"Reading %d bytes from file %s failed", size, fullpath);
		return false;
	}


	Loader_file *newfile = new Loader_file(
		filename, offset, this, size, Loader_file::MC_NONE
	);

	files->insert(std::pair<const char*,Loader_file *>(newfile->getname(),newfile));

	return true;
}


//-----------------------------------------------------------------------
//class Loader_sources: public std::list<Loader_source *>{};
//class Loader_files  : public std::hash_map<const char *,Loader_file *,hash<const char*>,Streq>{};

//============================================================================
//============================================================================
//----------------------------------------------------------------------------
Loader::Loader()
{
	sources = new Loader_sources;
	files = new Loader_files;

	//initialize the highlevel chaches:
	graphics.init(this);
	fonts.init(this);
	animations.init(this);
}

//----------------------------------------------------------------------------
Loader::~Loader()
{
	Loader_sources::iterator s= sources->begin();
	while (s != sources->end()) {
		delete *s;
		s++;
	}
	delete sources;
	
	if (!files->empty()) {
		logmsg(lt_warning,"After deleting all the data sources there were still some files left");
	}

	delete files;
}

typedef enum Sourcekind {
	sk_dir,		
	sk_file,
	sk_lump,
	sk_error
};

//----------------------------------------------------------------------------
Sourcekind getsourcekind(char *path)
{
	//first try for straight file or directory
	struct stat stats;
	if (stat(path,&stats) == 0) {
		if (S_ISREG(stats.st_mode))
			return sk_file;

		if (S_ISDIR(stats.st_mode))
			return sk_dir;
	}

	//then try for path.lump
	char *temp = (char *)malloc(strlen(path)+20);
	strcpy(temp,path);
	strcat(temp,".lump");

	if (stat(temp,&stats) == 0)
		if (S_ISREG(stats.st_mode)) {
			free(temp);
			return sk_lump;
		}
	free(temp);
	return sk_error;
}

//----------------------------------------------------------------------------
int Loader::addsources(const char *paths)
{
	char *tmp = strdup(paths);
	char *tmp_saved=tmp;

	char *path = strsep(&tmp,"\t");
	while (path && *path) {
		addsource(path);
		path = strsep(&tmp,"\t");
	}

	free(tmp_saved);

	return 1;
}

//----------------------------------------------------------------------------
int Loader::addsource(char *path)
{
	char expanded_path[200];
	if (!expand_path(path,expanded_path,sizeof(expanded_path))) {
		logmsg(lt_error,"Unable to expand the path %s",path);
		return -4;
	}

	int old_file_size = files->size();
	Sourcekind sourcekind = getsourcekind(expanded_path);
	switch (sourcekind) {
		case sk_dir: {
			Loader_source_directory *source = new Loader_source_directory;
			if (!source->load_directory(files,expanded_path)) {
				delete source;
				logmsg(lt_error,"Unable to load the directory %s",expanded_path);
				return -1;
			}

			sources->push_back(source);

			break;
		}

		case sk_lump:{
			strcat(expanded_path,".lump");
			//no, there is no break here.
		}

		case sk_file: {
			Loader_source_lump *source = new Loader_source_lump;
			if (!source->load_lump(files,expanded_path)) {
				delete source;
				logmsg(lt_error,"Unable to load the lump %s",expanded_path);
				return -1;
			}

			sources->push_back(source);

			break;
		}

		case sk_error:{
			logmsg(lt_error,"Unable to find the datasource %s",expanded_path);
			return -4;
			break;
		}
	}

	int file_size = files->size();
	int file_count = file_size-old_file_size;
	logmsg(lt_debug,"Found %i files at %s",file_count,expanded_path);
	
	return file_count;
}

//----------------------------------------------------------------------------
int Loader::filesize(char *name)
{
	Loader_files::iterator file = files->find(name);
	if (file == files->end()) {
		logmsg(lt_error,"Unable to find the file %s to get the file size",name);
		return -1;
	}
	
 	return file->second->getsize();
}

//----------------------------------------------------------------------------
bool Loader::fileread(char *name, void *buffer, int buffer_size, int offset)
{
	Loader_files::iterator file = files->find(name);
	if (file == files->end()) {
		logmsg(lt_error,"Unable to find the file %s for fileread()",name);
		return false;
	}

	return file->second->fileread(buffer, buffer_size, offset);
}


//----------------------------------------------------------------------------
const void *Loader::getfile(char *name)
{
	Loader_files::iterator file = files->find(name);
	if (file == files->end()) {
		logmsg(lt_error,"Unable to find the file %s for getfile(), the signs point to crashes.",name);
		return NULL;
	}

	return file->second->getdata();
}

//----------------------------------------------------------------------------
int Loader::save_lump(char *buffer, int buffer_size)
{
	//first make sure we at least have space for the basic stuff
	if ((unsigned)buffer_size < sizeof(Lump_header)+sizeof(Lump_directory_header)) {
		logmsg(lt_error,"Unable to save_lump, the buffer is only %i bytes long, this is way too small.",buffer_size);
		return -1;
	}

	//start writing just after the Lump_header hich we will write later.
	int current_pos = sizeof(Lump_header);

	int file_count = files->size();
	Lump_directory_item *items = (Lump_directory_item *)malloc(file_count*sizeof(Lump_directory_item));
	char **names = (char**)malloc(file_count*sizeof(char*));
	
	//write the files:
	int item_index = 0;
	Loader_files::iterator file = files->begin();
	while (file != files->end()) {
  	
		//pad with zeros so all files start at a 4 byte boundry
		while (current_pos & 3) {
			buffer[current_pos++] = 0;
		}

		//check that we have space for the file
		if (current_pos + file->second->getsize() + 4 > buffer_size) {
			logmsg(lt_error,"Unable to save_lump, ran out of space in the buffer.");
			free(items);
			free(names);
			return -2;
		}

		//collect the data for the directory:
		items[item_index].magic = LUMP_DIRECTORY_ITEM_MAGIC;
		items[item_index].file_offset = current_pos;
		items[item_index].file_length = file->second->getsize();
		items[item_index].name_length = strlen(file->second->getname())+1;
		items[item_index].swap();
		names[item_index] = (char *)file->second->getname();
		item_index++;

		//actually copy the data
		memcpy(buffer+current_pos,file->second->getdata(),file->second->getsize());
		current_pos += file->second->getsize();
		
		file++;
	}

	//write the header (at the start of the file)
	Lump_header header;
	header.magic = LUMP_HEADER_MAGIC;
	header.directory_offset = current_pos;
	header.swap();
	memcpy(buffer,&header,sizeof(header));

	//check we have room for the directory:
	if (current_pos + sizeof(Lump_directory_header) + file_count*sizeof(Lump_directory_item) > (unsigned)buffer_size) {
		logmsg(lt_error,"Unable to save_lump, ran out of space for the directory.");
		free(items);
		free(names);
		return -3;
	}
	
	//Write the directory
	Lump_directory_header directory;
	directory.magic = LUMP_DIRECTORY_HEADER_MAGIC;
	directory.items = file_count;
	directory.swap();
	memcpy(buffer+current_pos,&directory,sizeof(directory));
	current_pos+=sizeof(directory);

	//write the directory items:
	for (int i=0;i<file_count;i++) {

		memcpy(buffer+current_pos,items+i,sizeof(Lump_directory_item));
		current_pos += sizeof(Lump_directory_item);
		
		memcpy(buffer+current_pos,names[i],strlen(names[i])+1);
		current_pos += strlen(names[i])+1;
	}

	free(items);
	free(names);
 	return current_pos;
}
