/***************************************************************************
                          object.h  -  description
                             -------------------
    begin                : Fri Oct 8 1999
    copyright            : (C) 1999 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/

#ifndef BASEOBJECT_H
#define BASEOBJECT_H

//---------------------------------------------------------------------------
class World;
class Obj_player_avatar;
class Obj_player_data;

#include "fluff.h"
#include "nil_math.h"
#include "viewport.h"
#include "world.h"
#include "serverworld.h"
#include "serializable.h"

//Kinds of hit
typedef enum Hit_type {
	ht_none,    //didn't hit anything
	ht_object,	
	ht_map,			//we hit a pixel on the map
	ht_left,		//hit left side of the map
	ht_top,     //hit top side of the map
	ht_right,		//hit right side of the map
	ht_bottom		//hit bottom side of the map
};

typedef struct Map_hit {
	Hit_type hittype;
	int xc;
	int yc;	
	int objectid;
};


//--- ---------------------------------------------------------------
//All objects that are to be drawn must inherit from this class
//An object in the system
class Obj_base : public Serializable {
	public:
	Obj_base(World *world_, int id);
	virtual ~Obj_base();

/**********************************************************************
 *                  *Start of MANDATORY functions*
 *      These fuctions MUST be implemented by ALL clild objects
 **********************************************************************/

	/* Must return the type of the concrete class
	 */
	virtual Object_type gettype()=0;

	/* Does the thinking needed for the object on the serverside.
   * Thinking on the server means checking for collisions,
   * becoming an explosion and hurting players in range of explosions.
   * Calls client_think() for updating of animation the object and applying physics
	 */
 	virtual void server_think()=0;

	/* Draws the object into the viewport
   * This routine should handle clipping on its own.
	 */
	virtual void draw(Viewport *viewport)=0;

/**********************************************************************
 *                   *End of MANDATORY functions*
 **********************************************************************/

	/* called to read or write the state of the object
	 */
	virtual void serialize(Serializer *serializer);

	/* Does the thinking needed for the object on the client side (that is completely non-random)
   * Thinking means updating the physics and other deterministic things, such as animations.
   * MUST NOT alter other objects, this must be handled by server_think()
	 */
	virtual void client_think();

	/* return true if this object can be hurt, if it can it should also implement hurt()
   * Do you really want to hurt me?
	 */
	virtual bool hurtable() {return false;}

	/* called when something hurts the object,
	 * the default is to die if it is hurt at all, this is not very practical for most objects:)
	 */
	virtual void hurt(float hitpoints, Vector speed, int hurter) {if (hitpoints>0) done=true;}

	/*
   * Used for collisiondetection, alter it if you need to have another shape than a point.
	 */
	virtual bool is_passable(int offset) {return IS_PASSABLE(world->map_buffer.get_pixels()[offset]);}

	/* Be done, the object should make peace with it's maker and set the done flag...
	 */
	virtual void bedone() {done = true;}

	/* Returns true once the object wants to die
   * In the server this means that the object must be delete'ed
   * and that the clients should be told to do the same.
	 */
	virtual bool isdone() {return done;}

	//Misc access functions:

	/*
		The objectid uniquely indentifies all objects in the system,
		pointers should only be used inside a function as they cannot
		cross from the server to the client as the id's can
		and pointers have a way of leading to madness once the target object
		dies (something that happens quite often in this program)
	*/
	int getid() {return objectid;};

	//used while saving/sending state:
	bool isdirty() { return net_dirty;}
	void beclean() { net_dirty = false;}
  void setdirty() {net_dirty = true;};

	void getpos(float &x, float &y) {x = xpos; y=ypos;}
	void getspeed(float &xspeed_, float &yspeed_) {xspeed_=xspeed; yspeed_=yspeed;}
	void push(Vector &speed) {xspeed += speed.x; yspeed += speed.y; setdirty();}

	//Getting different kinds of "owners":
	int getowner() {return ownerid;}
	Obj_player_data *get_owner_data();
	Obj_player_avatar *get_owner_avatar();

	//collision helper:
	bool is_passable(int x, int y) {return is_passable(x+y*world->map_buffer.get_xsize());}

	//An easy way to get the name of an object
	const char *getname() {return OBJECT_NAMES[gettype()];}


	protected:

	/* Makes the object do a bogus bounce (not a bounce at all, it's an imitation)
	 */
	void bogo_bounce();

	/* Call from server_think to call client_think and check for collisions.
	 */
	Map_hit update_simple(bool check_object_collision = false);

  bool canmove();
	bool onground();
	
  World *world;
	
	//this object has had it's state updated in such a way that it should now be serialized and sent to the client
	bool net_dirty;
	bool done;

	//id of this object, assigned by the server at creation
	int objectid;	

	//---------------------------------------------------------------------------------
	//intenal helpers:

  //updates the common physical data according to Newtons laws (although me and hawking know that they are bollox)
	void newton_update(float newtime);
	float age() {return world->gettime()-createtime;}

 	  ///////////////////////////
	 // Syncronized data Start//
	///////////////////////////
	//position
	float xpos;
	float ypos;

	//speed
	float xspeed;
	float yspeed;

	//acceleration
	float yaccel;

	//This works against xspeed and yspeed, and also leads to a terminal speed when falling
	float air_resistance;

	//the time the object was created
	float createtime;

	//the time this data was last updated
	float curtime;

	//the owner of the object (there isn't always one)
	int ownerid;				//player id of the owner of this object

	int radius; 				//used by the collision routines <= 0 means that it should use is_passable() to test
 	  ///////////////////////////
	 // Syncronized data end  //
	///////////////////////////

	//non-synced data, that is simply good to have around on the server...
	float old_xpos;
	float old_ypos;
	float old_xspeed;
	float old_yspeed;
};

#endif

















































































