/***************************************************************************
                          raw_surface.cpp  -  description
                             -------------------
    begin                : Mon Jan 3 2000
    copyright            : (C) 2000 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/


#include "raw_surface.h"
#include "../../nil/nil/fluff.h"
#include "../../nil/nil/nil_math.h"

//-------------------------------------------------------------------
Raw_surface::Raw_surface()
{
	xsize = 0;
	ysize = 0;
	pixels = NULL;  	
	attached = false;
}

//-------------------------------------------------------------------
Raw_surface::~Raw_surface()
{
	free_pixels();
}

//-------------------------------------------------------------------
void Raw_surface::free_pixels()
{
	if (pixels && !attached) {
		free(pixels);
	}

	attached = false;
	xsize = 0;	
	ysize = 0;	
  pixels = NULL;
}


/****************************************************
	* Creating this object:
	*/

//-------------------------------------------------------------------
/** Create a surface that is managed by this object
	*/
void Raw_surface::create(int xsize_, int ysize_)
{
	free_pixels();

	if (!xsize_) return;
	if (!ysize_) return;

	xsize = xsize_;
	ysize = ysize_;
	pixels = (int32*)malloc(xsize*ysize*sizeof(int32));	
	attached = false;
}

//-------------------------------------------------------------------
void Raw_surface::create(Raw_surface *original)
{
	create(original->get_xsize(),original->get_ysize());
	memcpy(pixels,original->get_pixels(),original->get_xsize()*original->get_ysize()*sizeof(int32));
}

//-------------------------------------------------------------------
/** attach to a surface that is managed outside this object
	*/
void Raw_surface::attach(int xsize_, int ysize_, int32 *pixels_)
{
	free_pixels();

	if (!pixels_) return;
	if (!xsize_) return;
	if (!ysize_) return;

	xsize = xsize_;
	ysize = ysize_;
	pixels = pixels_;	
	attached = true;
}

#ifdef USE_PTC
//-------------------------------------------------------------------
/** attach to a PTC Surface (this will lock the PTC surface)
	*/
void Raw_surface::attach(BaseSurface *ptc_surface)
{
	free_pixels();

	if (!ptc_surface) return;

	pixels = (int32*)ptc_surface->lock();

	if (!pixels) return;
	
	xsize = ptc_surface->width();
	ysize = ptc_surface->height();	
	attached = true;
}
#endif

/****************************************************
	* These routines will draw on another Raw_surface:
	*/

//-------------------------------------------------------------------
void Raw_surface::draw_copy(Raw_surface *target,
			int source_xpos, int source_ypos,
			int target_xpos, int target_ypos,
			int target_xsize, int target_ysize)
{
	//Copy the pixels one line at the time
	int t_xsize = target->get_xsize();
	int bytes_pr_line = target_xsize*sizeof(int32);
	int32 *target_pixel = target->get_pixels()+source_xpos+ source_ypos*t_xsize;
	int32 *source_pixel = pixels+target_xpos+target_ypos*xsize;

	for(int y=0;y<target_ysize;y++) {
		memcpy(target_pixel,source_pixel,bytes_pr_line);
		target_pixel+=t_xsize;
		source_pixel+=xsize;
	}
}

//-------------------------------------------------------------------
void Raw_surface::draw_copy_stretch(Raw_surface *target,
			int source_xpos,  int source_ypos,
			int source_xsize, int source_ysize,
			int target_xpos,  int target_ypos,
			int enlargement)
{
	//Copy the pixels one line at the time
	int t_xsize = target->get_xsize();
	int32 *target_pixel = target->get_pixels()+source_xpos+ source_ypos*t_xsize;
	int32 *source_pixel = pixels+target_xpos+target_ypos*xsize;

	if (enlargement == 1) {

		int bytes_pr_line = source_xsize*sizeof(int32);
		for(int y=0;y<source_ysize;y++) {
			memcpy(target_pixel,source_pixel,bytes_pr_line);
			target_pixel+=t_xsize;
			source_pixel+=xsize;
		}

	} else if (enlargement == 2) {
		for(int y=0;y<source_ysize;y++) {

			int32 *sp = source_pixel;
			int32 *tp = target_pixel;
			{ //MS compiler bug workaround
				for(int x=0;x<source_xsize;x++) {
  				*(tp++) = *sp;
  				*(tp++) = *sp;
					sp++;
				}
			}
			target_pixel+=t_xsize;

			sp = source_pixel;
			tp = target_pixel;
			{ //MS compiler bug workaround
				for(int x=0;x<source_xsize;x++) {
  				*(tp++) = *sp;
  				*(tp++) = *sp;
					sp++;
				}
			}
			target_pixel+=t_xsize;

			source_pixel+=xsize;
		}

	} else {

		for(int y=0;y<source_ysize;y++) {
//		memcpy(target_pixel,source_pixel,bytes_pr_line);

			for (int e=0;e<enlargement;e++) {
				int32 *sp = source_pixel;
				int32 *tp = target_pixel;

				for(int x=0;x<source_xsize;x++) {
					switch (enlargement) {
  		  	  case 7: *(tp++) = *sp;
  			    case 6: *(tp++) = *sp;
  			    case 5: *(tp++) = *sp;
  			    case 4: *(tp++) = *sp;
	  		    case 3: *(tp++) = *sp;
  		  	  case 2: *(tp++) = *sp;
						default:
  			    case 1: *(tp++) = *sp;
					}
					sp++;
				}

				target_pixel+=t_xsize;
			}

			source_pixel+=xsize;
		}
	}








}

//-------------------------------------------------------------------
/** Render this to another Raw_surface, without handling the flags
	*/
void Raw_surface::draw_alpha(Raw_surface *target, int xpos, int ypos)
{
	if (!pixels) return;

	//clipping variables:
	int clipped_xstart = 0;
	int clipped_ystart = 0;

	if (xpos < 0)	clipped_xstart = -xpos;
	if (ypos < 0)	clipped_ystart = -ypos;

	int clipped_xsize = xsize-clipped_xstart;
	int clipped_ysize = ysize-clipped_ystart;

	int xsize_target = target->get_xsize(); //just a bit faster than using target->get_xsize() each time.
	if (xpos+xsize >= xsize_target)				clipped_xsize -= (xpos+xsize-xsize_target);
	if (ypos+ysize >= target->get_ysize())	clipped_ysize -= (ypos+ysize-target->get_ysize());

	int32 *target_line = target->get_pixels() + xpos+clipped_xstart + (ypos+clipped_ystart)*xsize_target;
	int32 *source_line = pixels               + clipped_ystart*xsize+ clipped_xstart;

	for (int yc=0; yc < clipped_ysize; yc++) {

		int32 *target = target_line;
		int32 *source = source_line;
		for (int xc=0; xc < clipped_xsize; xc++) {

			int32 s = *source;

    	if (IS_OPAQUE(s))
				*target = s;

			else if (!IS_INVISIBLE(s)) {
      	//ARGH! ARGH! it's not opaque and it's not invisible, so I gotta do alpha now.

				int32 t = *target;

				int a = GETALPHA(s);
				int a1 = 16-a;

				int r = (GETRED(s)*a   + GETRED(t)*a1)   >> 4;
				int g = (GETGREEN(s)*a + GETGREEN(t)*a1) >> 4;
				int b = (GETBLUE(s)*a  + GETBLUE(t)*a1)  >> 4;

				*target = PIXEL(r,g,b);				
			}
			
      target++;
			source++;
		}	

		target_line += xsize_target;
		source_line += xsize;		
	}
}

//-------------------------------------------------------------------
/** Render this to another Raw_surface, with viewport-centric handling of the flags
	*/
void Raw_surface::draw_view(Raw_surface *target, int xpos, int ypos)
{
	if (!pixels) return;

	//calc start coords of the sprite
	xpos -= xsize >> 1;
	ypos -= ysize >> 1;

	//clipping variables:
	int clipped_xstart = 0;
	int clipped_ystart = 0;

	if (xpos < 0)	clipped_xstart = -xpos;
	if (ypos < 0)	clipped_ystart = -ypos;

	int clipped_xsize = xsize-clipped_xstart;
	int clipped_ysize = ysize-clipped_ystart;

	int xsize_target = target->get_xsize(); //just a bit faster than using target->get_xsize() each time.
	if (xpos+xsize >= xsize_target)				clipped_xsize -= (xpos+xsize-xsize_target);
	if (ypos+ysize >= target->get_ysize())	clipped_ysize -= (ypos+ysize-target->get_ysize());

	int32 *target_line = target->get_pixels() + xpos+clipped_xstart + (ypos+clipped_ystart)*xsize_target;
	int32 *source_line = pixels                   + clipped_ystart*xsize    + clipped_xstart;

	for (int yc=0; yc < clipped_ysize; yc++) {

		int32 *target = target_line;
		int32 *source = source_line;
		for (int xc=0; xc < clipped_xsize; xc++) {

			int32 s = *source;

    	if (IS_OPAQUE(s))
				*target = s;

			else if (!IS_INVISIBLE(s)) {
      	//ARGH! ARGH! it's not opaque and it's not invisible, so I gotta do alpha now.

				int32 t = *target;

				int a = GETALPHA(s);
				int a1 = 16-a;

				int r = (GETRED(s)*a   + GETRED(t)*a1)   >> 4;
				int g = (GETGREEN(s)*a + GETGREEN(t)*a1) >> 4;
				int b = (GETBLUE(s)*a  + GETBLUE(t)*a1)  >> 4;

				*target = PIXEL(r,g,b) | PF_TYPE_EARTH;				
			}
			
      target++;
			source++;
		}	

		target_line += xsize_target;
		source_line += xsize;		
	}
}

//-------------------------------------------------------------------
/** Render this to another Raw_surface, with map-centric handling of the flags
	* This is the slowest of the drawing routines
	*/
void Raw_surface::draw_map(Raw_surface *target_map, Raw_surface *target_bgmap, int xpos, int ypos)
{
	if (!pixels) return;

	//calc start coords of the sprite
	xpos -= xsize >> 1;
	ypos -= ysize >> 1;

	//clipping variables:
	int clipped_xstart = 0;
	int clipped_ystart = 0;

	if (xpos < 0)	clipped_xstart = -xpos;
	if (ypos < 0)	clipped_ystart = -ypos;

	int clipped_xsize = xsize-clipped_xstart;
	int clipped_ysize = ysize-clipped_ystart;

	int xsize_target = target_map->get_xsize(); //just a bit faster than using raw_target->get_xsize() each time.
	if (xpos+xsize >= xsize_target)							clipped_xsize -= (xpos+xsize-xsize_target);
	if (ypos+ysize >= target_map->get_ysize())	clipped_ysize -= (ypos+ysize-target_map->get_ysize());

	int32 *target_line = target_map->get_pixels() + xpos+clipped_xstart + (ypos+clipped_ystart)*xsize_target;
	int32 *bgmap_line  = target_bgmap->get_pixels()  + xpos+clipped_xstart + (ypos+clipped_ystart)*xsize_target;
	int32 *source_line = pixels                   + clipped_ystart*xsize    + clipped_xstart;

	for (int yc=0; yc < clipped_ysize; yc++) {

		int32 *target = target_line;
		int32 *source = source_line;
		int32 *bgmap = bgmap_line;
		for (int xc=0; xc < clipped_xsize; xc++) {

			int32 s = *source;
			int32 t = *target;

			int32 type = s & PF_TYPE_BITS;
	
			if (IS_STONE(t)) { 								//1
				type = PF_TYPE_STONE;

			} else if (IS_HOLE(s) || IS_EARTH(s) || IS_STONE(s)) { //2
				type = s & PF_TYPE_BITS;

			} else {
      	type = t & PF_TYPE_BITS;				//3 (target type unchanged)
			}

			if (!IS_STONE(t)) {

				if (IS_HOLE(type)) {
					*target = *bgmap;
 	
				} else if (IS_EARTH(type) || IS_STONE(type)) {
 	
		    	if (IS_OPAQUE(s))
						*target = (s & PF_RGB_BITS) | type;
  	
					else if (IS_INVISIBLE(s)) {
						*target = (t & PF_RGB_BITS) | type;
   	
					}	else {
	  	    	//ARGH! ARGH! it's not opaque and it's not invisible, so I gotta do alpha now.
 	  	
						int a = GETALPHA(s);
						int a1 = 16-a;
  		
						int r = (GETRED(s)*a   + GETRED(t)*a1)   >> 4;
						int g = (GETGREEN(s)*a + GETGREEN(t)*a1) >> 4;
						int b = (GETBLUE(s)*a  + GETBLUE(t)*a1)  >> 4;
  		
						*target = PIXEL(r,g,b) | type;				
					}
				} //stone/earth

			} //skip put
			
      target++;
			source++;
			bgmap++;
		}	

		target_line += xsize_target;
		bgmap_line += xsize_target;
		source_line += xsize;		
	}	
}

/****************************************************
	* Loading and saving routines:
	*/
//-------------------------------------------------------------------
bool Raw_surface::save_to_file(char *filename)
{
	if (!pixels) return false;

	FILE *f=fopen(filename,"wc");
	if (!f) return false;

	Rgbf_header header;
	header.magic = RGBF_HEADER_MAGIC;
	header.xsize = xsize;
	header.ysize = ysize;
	header.swap();

	if (fwrite(&header,sizeof(header),1,f) != 1) {fclose(f); return false;}
	if (fwrite(pixels,xsize*ysize*sizeof(int32),1,f) != 1) {fclose(f); return false;}

	fclose(f);

	return true;
}

//-------------------------------------------------------------------
bool Raw_surface::load_from_file(char *filename)
{
	FILE *f = fopen(filename,"r");
	if (!f) {
		//complain to someone that the file could not be opened?
  	return false;
	}

	Rgbf_header header;
	if (fread(&header,sizeof(header),1,f) != 1) {
	  //complain to someone that we couldn't read the header?
   	fclose(f); 	
 		return false;
	}
	header.swap();

	if (header.magic != RGBF_HEADER_MAGIC) {
		//maybe log a message here.
		fclose(f);
		return false;
	}

	create(header.xsize,header.ysize);

	if (fread(pixels,xsize*ysize*sizeof(int32),1,f) != 1) {
		//maybe log a message here.
		free_pixels();
		fclose(f);
		return false;
	}

	fclose(f);

	return true;
}












/****************************************************
	* These routines will draw on this Raw_surface:
	*/

//-------------------------------------------------------------------
/** Draws a bresenham line with clipping
	*/
#define ERRORBITS 16
void Raw_surface::draw_line(int x0, int y0, int x1, int y1, int32 color, unsigned int style)
{
	if (!pixels) return;

	if (!clip_line(x0,y0,x1,y1)) return;

	//get delta values
	int dx = x1-x0;
	int dy = y1-y0;

	//zero length is boring so get rid of it.
	if (dx == 0 && dy == 0) return;

	if (ABS(dx) > ABS(dy)) {
		// x major
	
		int ymove = xsize*SIGN(dy);	//The step to take with y when error hits 1
		dy = (ABS(dy) << ERRORBITS)/ABS(dx);		//The value to add to error each x
		int e = 0;															//error, fixed point with errorbits bits between 0 and 1.
		int x = x0;															//x is just the x coordinate
		int y = y0*xsize;						//y is offset into buffer, not the y coordinate (this saves many muls)
		dx = SIGN(dx);													//we just need the sign of dx
		int old_y = y;

		x1 += dx;
		while (x != x1) {

			//set the pixel
			pixels[x+y] = color;

			//tally the error and ajust
			e += dy;
			if (e >= (1 << ERRORBITS)) {
				old_y = y;
				y += ymove;
				e -= (1 << ERRORBITS);
			}

			//move x one pixel in the right direction
			x += dx;

		} //while (x != x1) {
	
	} else {
		//y major

		int xmove = SIGN(dx);									//The step to take with x when error hits 1 << ERRORBITS
		dx = (ABS(dx) << ERRORBITS)/ABS(dy);	//The value to add to error each y
		int e = 0;														//error, fixed point with ERRORBITS bits between 0 and 1.
		int x = x0;														//x is just the x coordinate
		int y = y0*xsize;					//y is offset into buffer, not the y coordinate (this saves many muls)
		int sign_dy = SIGN(dy);
		dy = xsize*sign_dy;				//we just need the sign of dy scaled to one pixel up or down
		y1 = (y1+sign_dy)*xsize;
		int old_x = x;

		while (y != y1) {

			//set the color
			pixels[x+y] = color;

			//tally the error and ajust
			e += dx;
			if (e >= (1 << ERRORBITS)) {
				old_x = x;
				x += xmove;
				e -= (1 << ERRORBITS);
			}

			//move y one pixel in the right direction
			y += dy;

		} //while (y != y1) {

	}

}

//-------------------------------------------------------------------
/** draws a horiontal line (more quickly than the full bresenham routine)
  * Use this rather than draw_vertical_line() if practical, it is faster.
	*/
void Raw_surface::draw_horizontal_line(int x0, int y0, int x1, int32 color)
{
	if (!pixels) return;

	//Make sure that the x'es are in the correct order:
	if (x0 > x1) {
		int tmp = x1;
		x1 = x0;
		x0 = tmp;
	}

	//Do some clipping:
	if (y0 < 0 || y0 > ysize) return;
	x0 = CLAMP(x0,0,xsize);
	x1 = CLAMP(x1,0,xsize);
	if (x1 - x0 <= 0) return;

	//Then do the drawing:
	int32 *pixel = pixels+x0+y0*xsize;

	if (IS_OPAQUE(color)) {

	  for (int x=x0; x<x1; x++) {

			int32 pa = *pixel & PF_ALPHA_BITS;
			*pixel = color | pa;

			pixel++;
		}

	} else if (!IS_INVISIBLE(color)) {

		int a = GETALPHA(color);
		int a1 = 16-a;

		int cr = GETRED(color)*a;
		int cg = GETGREEN(color)*a;
		int cb = GETBLUE(color)*a;
		int32 ca = color & PF_ALPHA_BITS;

	  for (int x=x0; x<x1; x++) {

			int r = (cr + GETRED(*pixel)*a1)   >> 4;
			int g = (cg + GETGREEN(*pixel)*a1) >> 4;
			int b = (cb + GETBLUE(*pixel)*a1)  >> 4;

			*pixel = PIXEL(r,g,b) | ca;				

			pixel++;
		}
	}	
}


void Raw_surface::draw_horizontal_line_raw(int x0, int y0, int x1, int32 color)
{
	if (!pixels) return;

	//Make sure that the x'es are in the correct order:
	if (x0 > x1) {
		int tmp = x1;
		x1 = x0;
		x0 = tmp;
	}

	//Do some clipping:
	if (y0 < 0 || y0 > ysize) return;
	x0 = CLAMP(x0,0,xsize);
	x1 = CLAMP(x1,0,xsize);
	if (x1 - x0 <= 0) return;

	//Then do the drawing:
	int32 *pixel = pixels+x0+y0*xsize;
  for (int x=x0; x<x1; x++) {
		*pixel = color;
		pixel++;
	}
}

//-------------------------------------------------------------------
/** Draws a vertical line (more quickly than the full bresenham routine)
	*/
void Raw_surface::draw_vertical_line(int x0, int y0, int y1, int32 color)
{
	if (!pixels) return;

	//Make sure that the x'es are in the correct order:
	if (y0 > y1) {
		int tmp = y1;
		y1 = y0;
		y0 = tmp;
	}

	//Do some clipping:
	if (x0 < 0 || x0 > xsize) return;
	y0 = CLAMP(y0,0,ysize);
	y1 = CLAMP(y1,0,ysize);
	if (y1-y0 <= 0) return;

	//Then do the drawing:
	int32 *pixel = pixels+x0+y0*xsize;

	if (IS_OPAQUE(color)) {

	  for (int y=y0; y<y1; y++) {
			*pixel = color;
			pixel += xsize;
		}

	} else if (!IS_INVISIBLE(color)) {

		int a = GETALPHA(color);
		int a1 = 16-a;

		int cr = GETRED(color)*a;
		int cg = GETGREEN(color)*a;
		int cb = GETBLUE(color)*a;
		int32 ca = color & PF_ALPHA_BITS;

	  for (int y=y0; y<y1; y++) {

			int r = (cr + GETRED(*pixel)*a1)   >> 4;
			int g = (cg + GETGREEN(*pixel)*a1) >> 4;
			int b = (cb + GETBLUE(*pixel)*a1)  >> 4;

			*pixel = PIXEL(r,g,b) | ca;				

			pixel += xsize;
		}
	}	
}

void Raw_surface::draw_vertical_line_raw(int x0, int y0, int y1, int32 color)
{
	if (!pixels) return;

	//Make sure that the x'es are in the correct order:
	if (y0 > y1) {
		int tmp = y1;
		y1 = y0;
		y0 = tmp;
	}

	//Do some clipping:
	if (x0 < 0 || x0 > xsize) return;
	y0 = CLAMP(y0,0,ysize);
	y1 = CLAMP(y1,0,ysize);
	if (y1-y0 <= 0) return;

	//Then do the drawing:
	int32 *pixel = pixels+x0+y0*xsize;
  for (int y=y0; y<y1; y++) {
		*pixel = color;
		pixel += xsize;
	}
}

//-------------------------------------------------------------------
/** Draws a transparent rectangle, this is somewhat faster than using the draw_alpha() routine
	*/
void Raw_surface::draw_rectangle(int x0, int y0, int x1, int y1, int32 color)
{
	if (!pixels) return;

	if (y0 > y1) {
		int tmp = y1;
		y1 = y0;
		y0 = tmp;
	}

	if (x0 > x1) {
		int tmp = x1;
		x1 = x0;
		x0 = tmp;
	}

  if (x0 < 0) x0 = 0;
  if (y0 < 0) y0 = 0;
  if (x0 > xsize-1) return;
  if (y0 > ysize-1) return;

  if (x1 < 0) return;
  if (y1 < 0) return;
  if (x1 > xsize-1) x1 = xsize-1;
  if (y1 > ysize-1) y1 = ysize-1;
	
	int xcount = x1-x0;
	int ycount = y1-y0;

	//This is a speed hack so we favour horizontal lines
 	if (xcount > ycount/2) {
    for(int y=0;y<ycount;y++)
			draw_horizontal_line(x0,y0+y,x0+xcount,color);
	} else {
    for(int x=0;x<xcount;x++)
			draw_vertical_line(x0+x,y0,y0+ycount,color);
	}
}


void Raw_surface::draw_rectangle_raw(int x0, int y0, int x1, int y1, int32 color)
{
	if (!pixels) return;

	if (y0 > y1) {
		int tmp = y1;
		y1 = y0;
		y0 = tmp;
	}

	if (x0 > x1) {
		int tmp = x1;
		x1 = x0;
		x0 = tmp;
	}

  if (x0 < 0) x0 = 0;
  if (y0 < 0) y0 = 0;
  if (x0 > xsize-1) return;
  if (y0 > ysize-1) return;

  if (x1 < 0) return;
  if (y1 < 0) return;
  if (x1 > xsize-1) x1 = xsize-1;
  if (y1 > ysize-1) y1 = ysize-1;
		
	int xcount = x1-x0;
	int ycount = y1-y0;

	//This is a speed hack so we favour horizontal lines
 	if (xcount > ycount/2) {
    for(int y=0;y<ycount;y++)
			draw_horizontal_line_raw(x0,y0+y,x0+xcount,color);
  } else {
    for(int x=0;x<xcount;x++)
			draw_vertical_line_raw(x0+x,y0,y0+ycount,color);
  }
}


//-------------------------------------------------------------------
/** Draws a rectangle (just the lines, not filled)
	*/
void Raw_surface::draw_rectangle_border(int x0, int y0, int x1, int y1, int32 color)
{
	if (!pixels) return;

	draw_horizontal_line(x0+1,y0,x1,color);
	draw_horizontal_line(x0,y1,x1+1,color);
	draw_vertical_line(x0,y0,y1,color);
	draw_vertical_line(x1,y0,y1,color);
}

void Raw_surface::draw_rectangle_border_raw(int x0, int y0, int x1, int y1, int32 color)
{
	if (!pixels) return;

	draw_horizontal_line_raw(x0+1,y0,x1,color);
	draw_horizontal_line_raw(x0,y1,x1+1,color);
	draw_vertical_line_raw(x0,y0,y1,color);
	draw_vertical_line_raw(x1,y0,y1,color);
}

//-------------------------------------------------------------------
void Raw_surface::draw_pixel(int x, int y, int32 color)
{
	if (x > 0 && x < xsize &&	y > 0 && y < ysize) {
		int32 *pixel = pixels + x+y*xsize;
		*pixel = color;
	}
}

//-------------------------------------------------------------------
void Raw_surface::draw_pixel_special(int x, int y, int32 color)
{
	if (x > 0 && x < xsize &&	y > 0 && y < ysize) {
		int32 *pixel = pixels + x+y*xsize;

		if (!IS_PASSABLE(*pixel)) return;

  	int32 type = *pixel | PF_TYPE_BITS;
	
   	if (IS_OPAQUE(color))
			*pixel = (*pixel & PF_RGB_BITS) | type;

		else if (IS_INVISIBLE(color)) {
			//Don't modify the pixel at all
   	
		}	else {
    	//ARGH! ARGH! it's not opaque and it's not invisible, so I gotta do alpha now.
 	  	
			int a = GETALPHA(color);
			int a1 = 16-a;
  		
			int r = (GETRED(color)*a   + GETRED(*pixel)*a1)   >> 4;
			int g = (GETGREEN(color)*a + GETGREEN(*pixel)*a1) >> 4;
			int b = (GETBLUE(color)*a  + GETBLUE(*pixel)*a1)  >> 4;
  		
			*pixel = PIXEL(r,g,b) | type;				
		}
	}
}

//-------------------------------------------------------------------
void Raw_surface::modify_pixel_type(int x, int y, int32 type_flags)
{
	if (x > 0 && x < xsize &&	y > 0 && y < ysize) {
		int32 *pixel = pixels + x+y*xsize;
		*pixel = (PF_RGB_BITS & *pixel) | type_flags;
	}
}

/****************************************************
	* These routines are helpers for others, they may be of limited use to most people
	*/

//-------------------------------------------------------------------
/** Use this in line drawing routines to have the target surface clip the line for the source
	* returns false if the line is not visible at all
	*/
bool Raw_surface::clip_line(
		int &x0, int &y0, 										
		int &x1, int &y1
	)
{
	if (!pixels) return false;

  Area_code code0 = area_code(x0,y0);
  Area_code code1 = area_code(x1,y1);

  while (code0 != AC_INSIDE || code1 != AC_INSIDE) {
		// While both ends are not inside the clipping region
  	
		if (code0 & code1) {
			return false; // Trivial Reject (each end is off to the same side)

    } else {

			// Failed both tests, so calculate the line segment to clip
  		Area_code codeout;

      if (code0)
        codeout = code0; // Clip the first point
      else
        codeout = code1; // Clip the last point

  		int x=0;
			int y=0;

      if (codeout & AC_BOTTOM) {
				// Clip the line to the bottom of the viewport
        y = ysize-1;
        x = x0+(x1-x0)*(y-y0) / (y1 - y0);

      } else if (codeout & AC_TOP) {
      	// Clip the line to the top of the viewport
        y = 0;
        x = x0+(x1-x0)*(y-y0) / (y1 - y0);

      } else if (codeout & AC_RIGHT) {
				// Clip the line to the right edge of the viewport
        x = xsize-1;
        y = y0+(y1-y0)*(x-x0) / (x1-x0);

      } else if (codeout & AC_LEFT) {
        // Clip the line to the left edge of the viewport
        x = 0;
        y = y0+(y1-y0)*(x-x0) / (x1-x0);
      }

      if (codeout == code0) {
				// Modify the first coordinate
        x0 = x; y0 = y;
        code0 = area_code(x0, y0);

      } else {
      	//Modify the second coordinate
        x1 = x; y1 = y;
        code1 = area_code(x1, y1);
      }

    } // if (!(code0 & code1))

  } // while (code0 <> 0 || code1 <> 0) do {
	return true;
}

//Figure out what area code a point is in
Raw_surface::Area_code Raw_surface::area_code(int x, int y)
{        			
	Area_code code = AC_INSIDE;

  if (y > ysize-1) 	code = AC_BOTTOM;
  else if (y < 0) 	code = AC_TOP;

  if (x > xsize-1) 	code |= AC_RIGHT;
  else if (x < 0) 	code |= AC_LEFT;

  return code;
}

