/***************************************************************************
                          raw_surface.h  -  description
                             -------------------
    begin                : Mon Jan 3 2000
    copyright            : (C) 2000 by Flemming Frandsen
    email                : dion@swamp.dk

  *@author Flemming Frandsen
	* This class used to store the lowlevel graphics rendering routines
	* Note that pixels are ALWAYS 32 bit and in the NiL format

 ***************************************************************************/


#ifndef RAW_SURFACE_H
#define RAW_SURFACE_H

#include "../../nil/nil/nil_files.h"

#ifdef USE_PTC
#include "ptc/ptc.h"
#endif

//this must always become a 32bit integer:
typedef unsigned int Pixel;

//emulate ptc names:
typedef Pixel int32;

//---------------------------------------------------------
//pixelflags used in the map:
#define PF_TYPE_BIT0	0x01000000 	
#define PF_TYPE_BIT1	0x02000000
#define PF_TYPE_BITS	(PF_TYPE_BIT1 | PF_TYPE_BIT0)

//unused:
#define PF_FLAG2 0x04000000
#define PF_FLAG3 0x08000000

//Alpha:
#define PF_ALPHA_BIT0 0x10000000
#define PF_ALPHA_BIT1 0x20000000
#define PF_ALPHA_BIT2 0x40000000
#define PF_ALPHA_BIT3 0x80000000
#define PF_ALPHA_BITS 0xF0000000

//RGB
#define PF_RGB_BITS	  0x00FFFFFF

//---------------------------------------------------------
// Meaning of the pixel types bits:
#define PF_TYPE_VOID 0
#define PF_TYPE_HOLE PF_TYPE_BIT0
#define PF_TYPE_EARTH PF_TYPE_BIT1
#define PF_TYPE_STONE (PF_TYPE_BIT0 | PF_TYPE_BIT1)

//---------------------------------------------------------
//Define a couple of macros to access the pixel data
#define GETFLAG(x)   ((x) & 0xFF000000)
#define GETRED(x)   (((x) & 0x00FF0000) >> 16)
#define GETGREEN(x) (((x) & 0x0000FF00) >> 8)
#define GETBLUE(x)  (((x) & 0x000000FF))

#define GETALPHA(x) ((x) >> (32-4))

//pixel tests:
#define IS_VOID(x)	(((x) & PF_TYPE_BITS) == PF_TYPE_VOID)
#define IS_HOLE(x)	(((x) & PF_TYPE_BITS) == PF_TYPE_HOLE)
#define IS_EARTH(x)	(((x) & PF_TYPE_BITS) == PF_TYPE_EARTH)
#define IS_STONE(x)	(((x) & PF_TYPE_BITS) == PF_TYPE_STONE)

#define IS_PASSABLE(x) (((x) & PF_TYPE_BITS) == PF_TYPE_HOLE)

#define IS_INVISIBLE(x) (((x) & PF_ALPHA_BITS) == 0x00000000)
#define IS_OPAQUE(x) (((x) & PF_ALPHA_BITS) == 0xf0000000)

#define PIXEL(r,g,b) ( (((r) & 0xff)<<16) | (((g) & 0xff)<<8) | ((b) & 0xff) )
#define PIXELA(r,g,b,a) ((((a) & 0x0f)<<28) | (((r) & 0xff)<<16) | (((g) & 0xff)<<8) | ((b) & 0xff) )

//---------------------------------------------------------
#define RGBF_HEADER_MAGIC_OLD 0xdeadbeef
typedef struct Rgbf_header_old {
	unsigned int magic; //set to 0xDEADBEEF
	int xsize;
	int ysize;
};


//---------------------------------------------------------
//returns the pixel format that is in use
#define PIXELFORMAT Format(32,0x00FF0000,0x0000FF00,0x000000FF)

/*
typedef union Pixel {
	unsigned int i;
  struct {
  	unsigned char r;
  	unsigned char g;
  	unsigned char b;
  	unsigned char f;
	}
}
*/
class Raw_surface {
public:
	Raw_surface();
	~Raw_surface();


	/** Create a surface that is managed by this object
  	*/
	void create(int xsize_, int ysize_);

	/** Create a surface by copying another
  	*/
	void create(Raw_surface *original);

	/** attach to a surface that is managed outside this object
  	*/
	void attach(int xsize_, int ysize_, int32 *pixels_);

#ifdef USE_PTC
	/** attach to a PTC Surface (this will lock the PTC surface)
  	*/
	void attach(BaseSurface *ptc_surface);
#endif


	/****************************************************
		* These routines will draw on another Raw_surface:
  	*/

	void draw_copy_stretch(Raw_surface *target,
			int source_xpos,  int source_ypos,
			int source_xsize, int source_ysize,
			int target_xpos,  int target_ypos,
			int enlargement = 1);


	/** Copy a rectangle to another Raw_surface, without handling flags or alpha (byte for byte copy)
  	*/
	void draw_copy(Raw_surface *target,
			int source_xpos, int source_ypos,
			int target_xpos, int target_ypos,
			int target_xsize, int target_ysize);

	/** Render this to another Raw_surface, without handling the flags
  	*/
	void draw_alpha(Raw_surface *target, int xpos, int ypos);

	/** Render this to another Raw_surface, with viewport-centric handling of the flags
  	*/
	void draw_view(Raw_surface *target, int xpos, int ypos);

	/** Render this to another Raw_surface, with map-centric handling of the flags
		* This is the slowest of the drawing routines
  	*/
	void draw_map(Raw_surface *target_map, Raw_surface *target_bgmap, int xpos, int ypos);


	/****************************************************
		* Loading and saving routines:
  	*/
	bool save_to_file(char *filename);
	bool load_from_file(char *filename);	

	/****************************************************
		* These routines are what most people will use to access the actual pixels:
  	*/

	/** Use this to access the image directly:
		*/
	int32 *get_pixels() {return pixels;}
	int get_xsize()			{return xsize;}
	int get_ysize()			{return ysize;}

	/****************************************************
		* These routines will draw on this Raw_surface:
    * The _raw versions will draw the color directly,
    * without regard to alpha, the normal ones will alphablend
  	*/

	/** Draws a bresenham line with clipping
		*/
	void draw_line(int x0, int y0, int x1, int y1, int32 color, unsigned int style=0);


	/** draws a horiontal line (more quickly than the full bresenham routine)
    * Use this rather than draw_vertical_line() if practical, it is faster.
		*/
	void draw_horizontal_line(int x0, int y0, int x1, int32 color);
	void draw_horizontal_line_raw(int x0, int y0, int x1, int32 color);

	/** Draws a vertical line (more quickly than the full bresenham routine)
		*/
	void draw_vertical_line(int x0, int y0, int y1, int32 color);
	void draw_vertical_line_raw(int x0, int y0, int y1, int32 color);

	/** Draws a transparent rectangle, this is somewhat faster than using the draw_alpha() routine
		*/
	void draw_rectangle(int x0, int y0, int x1, int y1, int32 color);
	void draw_rectangle_raw(int x0, int y0, int x1, int y1, int32 color);

	/** Draws a rectangle (just the lines, not filled)
		*/
	void draw_rectangle_border(int x0, int y0, int x1, int y1, int32 color);
	void draw_rectangle_border_raw(int x0, int y0, int x1, int y1, int32 color);

	/** Draws a single pixel (DO NOT USE THIS IF YOU NEED TO SET MORE THAN ONE (1) PIXEL!!)
		* This is as fast as possible for a single pixel, but horribly slow for multible pixels
		* as it has to do range checking and a mul for each pixel
		*/
	void draw_pixel(int x, int y, int32 color);

	//This handles alpha,
	void draw_pixel_special(int x, int y, int32 color);

	/** Modifies a the type of a single pixel, slowly
		*/
	void modify_pixel_type(int x, int y, int32 type_flags);

	/****************************************************
		* These routines are helpers for others, they may be of limited use to most people
  	*/

	/** Use this in line drawing routines to have the target surface clip the line for the source
		* returns false if the line is not visible at all
		*/
	bool clip_line(
			int &x0, int &y0, 										
			int &x1, int &y1
		);

protected:
	typedef int Area_code;
	enum {
		AC_INSIDE = 0,
		AC_BOTTOM = 1,
		AC_TOP    = 2,
		AC_LEFT   = 4,
		AC_RIGHT  = 8,
	};

	Area_code area_code(int x, int y);


	//All the data needed for a bitmap:
	int xsize;
	int ysize;
	int32 *pixels; 	

	bool attached; //if true, don't free the buffer

	//utility functions:
	void free_pixels();
};

#endif

