/***************************************************************************
                          tcp_server.cpp  -  description
                             -------------------
    begin                : Sat Nov 13 1999
    copyright            : (C) 1999 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "tcp_server.h"
#include "systemheaders.h"

//----------------------------------------------------------------------------
Tcp_server::Tcp_server()
{
	server_socket = 0;
	connections_inuse = 0;
	net_normal_backlog 		= 4;  	//Everything is ok as long as you don't get further behind than this
	net_abnormal_backlog 	= 20; 	//Kick clients that are this far behind.
	net_max_wait_time 		= 200; 	//the number of deca ms we wait for all clients to catch up.
}

void Tcp_server::set_fudge(int normal_backlog, int abnormal_backlog, int max_wait_time)
{
	net_normal_backlog 		= normal_backlog;
	net_abnormal_backlog 	= abnormal_backlog; 	
	net_max_wait_time 		= max_wait_time; 	
}

//----------------------------------------------------------------------------
Tcp_server::~Tcp_server()
{
	for (int c=0;c<connections_inuse;c++)
		delete connections[c];
	if (server_socket)
	  shutdown(server_socket,0);		
}

//----------------------------------------------------------------------------
int Tcp_server::bind_and_listen(int port, bool multithreaded_)
{
	multithreaded = multithreaded_;

	//create the socket
	server_socket = socket(AF_INET, SOCK_STREAM, getprotobyname("tcp")->p_proto);
	if (server_socket <= 0) {
  	return -1;
	}

	//bind to an addy
 	sockaddr_in serveraddr;
	serveraddr.sin_family = AF_INET;
  serveraddr.sin_port 	= htons(port);
  serveraddr.sin_addr.s_addr = INADDR_ANY;
	int bind_error = bind(server_socket, (sockaddr *) &serveraddr, sizeof(serveraddr));
	if (bind_error) {
  	return -2;
	}

	//start listening for messages:
	int listen_error = listen(server_socket,10);
	if (listen_error) {
  	return -3;
	}

	//disable nagle:
  int flag = 1;
  int opt_error = setsockopt(server_socket, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int));
	if (opt_error) {
  	return -4;
	}

	return 0;
}

//----------------------------------------------------------------------------
void Tcp_server::poll()
{
	//clean up after the connections that have died
	int put = 0;
	for (int c=0;c<connections_inuse;c++) {
		if (connections[c]->isdone())
			delete connections[c];
		else
			connections[put++] = connections[c];
	}
	connections_inuse = put;

	//keep accepting connections until there are no more in the queue
	bool connection_pending = false;
	do {
		fd_set rfds;
  	FD_ZERO(&rfds);
	  FD_SET(server_socket, &rfds);

	  timeval tv;
  	tv.tv_sec = 0;
	  tv.tv_usec = 0;

		int retval = select(server_socket+1, &rfds, NULL, NULL, &tv);
	  connection_pending = (retval > 0) && FD_ISSET(server_socket,&rfds);

		if (connection_pending) {
			logmsg(lt_debug,"going to accept connection");
			Tcp_server_connection *newconnection = new Tcp_server_connection;
			int new_error = newconnection->accept(this,server_socket);
			if (new_error == 0)	connections[connections_inuse++] = newconnection;
		}

	} while (connection_pending);

	//attempt to slow down the server until it holds a sensible framerate.
	int wait_time = 0;
	int waitcount = 0;
	bool all_ready;
	do {

		all_ready = true;
		for (int c=0;c<connections_inuse;c++) {
			if (!multithreaded)	connections[c]->poll();

			int latest_id = update_queue.get_latest_update_id();
			int ack_id = connections[c]->get_ack_update_id();
			int client_backlog =  latest_id-ack_id;

			//check to see if the client is down with the groove
			if (client_backlog > net_normal_backlog)
				all_ready = false;

			//check to see if the client is totally out of whack
			if (client_backlog > net_abnormal_backlog) {
				logmsg(lt_message,"A player fell too far behind (%i)and was kicked as a result. %i %i",client_backlog,ack_id,latest_id);
				if (connections[c]->player)	connections[c]->player->bedone();
				connections[c]->terminate();
			}
		} //for (int c=0;c<connections_inuse;c++) {

		if (!all_ready) {
			int ms = 5+waitcount++;
			usleep(ms*1000);
			wait_time += ms;
		}

	} while (!all_ready && wait_time < net_max_wait_time);
}

//----------------------------------------------------------------------------
//This should be called from time to time to find
//out what the oldest relevant update is and then have the update_queue
//recycle all serializers older than this.
void Tcp_server::recycle_serializers()
{
	int oldest_update_id = 1<<30;
	for (int i=0;i<connections_inuse;i++) {
		int current_update_id = connections[i]->get_current_update_id();
		if (current_update_id < oldest_update_id)
			oldest_update_id = current_update_id;
	}
	update_queue.recycle(oldest_update_id);
}
